@@ -29,6 +29,9 @@ use constant {
     _OP_FILE    => 3,
     _OP_LABEL   => 4,
     _OP_COMMENT => 5,
+
+    _FOR_LOOP   => 1,
+    _WHILE_LOOP => 2,
 };
 
 
@@ -90,6 +93,15 @@ my %unary = (
     'max_index' => 'max_index', # for loop context vars
 );
 
+my %goto_family = map { $_ => undef } qw(
+    for_iter
+    and
+    dand
+    or
+    dor
+    goto
+);
+
 my %builtin = (
     'raw'          => 'builtin_mark_raw',
     'html'         => 'builtin_html_escape',
@@ -173,12 +185,6 @@ has parser => (
     init_arg => undef,
 );
 
-has file => (
-    is  => 'rw',
-
-    init_arg => undef,
-);
-
 has input_layer => (
     is      => 'ro',
     default => ':utf8',
@@ -219,6 +225,12 @@ has current_file => (
     init_arg => undef,
 );
 
+has file => (
+    is  => 'rw',
+
+    init_arg => undef,
+);
+
 sub lvar_use {
     my($self, $n) = @_;
 
@@ -239,6 +251,7 @@ sub compile {
     local $self->{lvar_id     } = 0;
     local $self->{lvar}         = {};
     local $self->{const}        = [];
+    local $self->{in_loop}      = 0;
     local $self->{dependencies} = [];
     local $self->{cascade};
     local $self->{header}       = $self->{header};
@@ -597,7 +610,9 @@ sub _generate_print {
 
     foreach my $arg(@{ $node->first }){
         if(exists $Text::Xslate::OPS{$proc . '_s'} && $arg->arity eq 'literal'){
-            push @code, $self->opcode( $proc . '_s' => $arg->value );
+            push @code,
+                $self->opcode( $proc . '_s' => $arg->value,
+                               line         => $arg->line );
         }
         elsif($self->_can_print_optimize($proc, $arg)){
             my $filter      = $arg->first;
@@ -632,6 +647,7 @@ sub _generate_include {
             : $self->compile_ast($file) ),
         $self->opcode( $node->id => undef, line => $node->line ),
     );
+
     if(defined(my $vars = $node->second)) {
         @code = ($self->opcode('enter'),
             $self->_localize_vars($vars),
@@ -664,12 +680,13 @@ sub _generate_cascade {
     return;
 }
 
-sub _compile_block {
+# XXX: need more consideration
+sub _compile_loop_block {
     my($self, $block) = @_;
     my @block_code = $self->compile_ast($block);
 
     foreach my $op(@block_code) {
-        if($op->[_OP_NAME] eq 'pushmark') {
+        if(any_in( $op->[_OP_NAME], qw(pushmark loop_control))) {
             # pushmark ... funcall (or something) may create mortal SVs
             # so surround the block with ENTER and LEAVE
             unshift @block_code, $self->opcode('enter');
@@ -678,6 +695,18 @@ sub _compile_block {
         }
     }
 
+    foreach my $i(1 .. (@block_code-1)) {
+        my $op = $block_code[$i];
+        if($op->[_OP_NAME] eq 'loop_control') {
+            my $type = $op->[_OP_ARG];
+            $op->[_OP_NAME] = 'goto';
+
+            $op->[_OP_ARG] = (@block_code - $i);
+
+            $op->[_OP_ARG] += 1 if $type eq 'last';
+        }
+    }
+
     return @block_code;
 }
 
@@ -692,6 +721,7 @@ sub _generate_for {
     }
     local $self->{lvar}  = { %{$self->lvar} }; # new scope
     local $self->{const} = [];                 # new scope
+    local $self->{in_loop} = _FOR_LOOP;
 
     my @code = $self->compile_ast($expr);
 
@@ -700,13 +730,14 @@ sub _generate_for {
     my $lvar_name = $iter_var->id;
 
     $self->lvar->{$lvar_name} = $lvar_id;
+    $self->lvar->{'($_)'}     = $lvar_id;
 
     push @code, $self->opcode( for_start => $lvar_id, symbol => $iter_var );
 
     # a for statement uses three local variables (container, iterator, and item)
     local $self->{lvar_id} = $self->lvar_use(3);
 
-    my @block_code = $self->_compile_block($block);
+    my @block_code = $self->_compile_loop_block($block);
     push @code,
         $self->opcode( literal_i => $lvar_id, symbol => $iter_var ),
         $self->opcode( for_iter  => scalar(@block_code) + 2 ),
@@ -716,6 +747,26 @@ sub _generate_for {
     return @code;
 }
 
+sub _generate_for_else {
+    my($self, $node) = @_;
+
+    my $for_block  = $node->first;
+    my $else_block = $node->second;
+
+    my @code = (
+        $self->compile_ast($for_block),
+    );
+
+    # 'for' block sets __a with true if the loop count > 0
+    my @else = $self->compile_ast($else_block);
+    push @code, (
+        $self->opcode( or => scalar(@else) + 1, comment => 'for-else' ),
+        @else,
+    );
+
+    return @code;
+}
+
 sub _generate_while {
     my($self, $node) = @_;
     my $expr  = $node->first;
@@ -728,8 +779,10 @@ sub _generate_while {
 
     (my $cond_op, undef, $expr) = $self->_prepare_cond_expr($expr);
 
+    # TODO: combine all the loop contexts into single one
     local $self->{lvar}  = { %{$self->lvar}  }; # new scope
     local $self->{const} = [ @{$self->const} ]; # new scope
+    local $self->{in_loop} = _WHILE_LOOP;
 
     my @code = $self->compile_ast($expr);
 
@@ -744,7 +797,7 @@ sub _generate_while {
     }
 
     local $self->{lvar_id} = $self->lvar_use(scalar @{$vars});
-    my @block_code = $self->_compile_block($block);
+    my @block_code = $self->_compile_loop_block($block);
     return @code,
         $self->opcode( $cond_op => scalar(@block_code) + 2, symbol => $node ),
         @block_code,
@@ -753,6 +806,38 @@ sub _generate_while {
     return @code;
 }
 
+sub _generate_loop_control {
+    my($self, $node) = @_;
+    my $type = $node->id;
+
+    any_in($type, qw(last next))
+        or $self->_error("Oops: unknown loop control statement '$type'");
+
+    if(not $self->{in_loop}) {
+        $self->_error("Use of loop control statement ($type) outside of loops");
+    }
+
+    my @cleanup;
+    if( $self->{in_loop} == _FOR_LOOP && $type eq 'last' ) {
+        my $lvar_id = $self->lvar->{'($_)'};
+        defined($lvar_id)
+            or $self->_error('Oops: undefined loop iterator');
+
+        @cleanup = (
+            $self->opcode( 'nil', undef,
+                comment => 'to clean the loop context' ),
+            $self->opcode( save_to_lvar => $lvar_id + 0), # item
+            $self->opcode( save_to_lvar => $lvar_id + 1), # iterator
+            $self->opcode( save_to_lvar => $lvar_id + 2), # body
+            $self->opcode( literal_i    => 1 ), # for 'for-else'
+        );
+    }
+
+    return $self->opcode('leave'),
+           @cleanup,
+           $self->opcode('loop_control' => $type, comment => $type);
+}
+
 sub _generate_proc { # definition of macro, block, before, around, after
     my($self, $node) = @_;
     my $type   = $node->id;
@@ -938,10 +1023,10 @@ sub _generate_variable {
     }
 }
 
-sub _generate_marker {
+sub _generate_super {
     my($self, $node) = @_;
 
-    return return $self->opcode( $node->id => undef, symbol => $node );
+    return return $self->opcode( super => undef, symbol => $node );
 }
 
 sub _generate_literal {
@@ -1128,9 +1213,13 @@ sub _generate_iterator {
             $node);
     }
 
-    return $self->opcode( load_lvar => $lvar_id+1, symbol => $node );
+    return $self->opcode(
+        load_lvar => $lvar_id + 1,
+        symbol    => $node,
+    );
 }
 
+# $~iterator.body
 sub _generate_iterator_body {
     my($self, $node) = @_;
 
@@ -1141,7 +1230,10 @@ sub _generate_iterator_body {
             $node);
     }
 
-    return $self->opcode( load_lvar => $lvar_id+2, symbol => $node );
+    return $self->opcode(
+        load_lvar => $lvar_id + 2,
+        symbol    => $node,
+    );
 }
 
 sub _generate_assign {
@@ -1203,6 +1295,17 @@ sub _localize_vars {
     my($self, $vars) = @_;
     my @localize;
     my @pairs = @{$vars};
+
+    if( (@pairs % 2) != 0 ) {
+        if(@pairs == 1) {
+            return $self->compile_ast(@pairs),
+                $self->opcode( 'localize_vars' );
+        }
+        else {
+            $self->_error("You must pass pairs of expressions to include");
+        }
+    }
+
     while(my($key, $expr) = splice @pairs, 0, 2) {
         if(!any_in($key->arity, qw(literal variable))) {
             $self->_error("You must pass a simple name to localize variables", $key);
@@ -1271,15 +1374,6 @@ sub _fold_constants {
     return 1;
 }
 
-my %goto_family;
-@goto_family{qw(
-    for_iter
-    and
-    dand
-    or
-    dor
-    goto
-)} = ();
 
 sub _noop {
     my($self, $op) = @_;
